#include <cstdio>
#include <GL/glfw.h>
#include <FTGLTextureFont.h>
#include <fstream>
#include <vector>
#include <map>
#include <fmod.h>
#include <cstring>
#include "world.hpp"
#include "main.hpp"
#include "menus.hpp"

FTGLTextureFont font("gfx/Arialbd.ttf");

using namespace std;

void setupForWorld();
void setupForHUD();
void drawGameHUD();

bool exitGame = false;
bool showFps = false;
bool helpScreen = false;
bool beatenGame = false;
bool invert = false;
int camera = 1;
bool gotSound = false;
bool cpufriendly = false;

int curlev = 0;
vector<string> levels;
vector<string> levelnames;

void GLFWCALL keyHandler(int key, int action)
{
	if (action == GLFW_RELEASE)
		return;
	if (key == 'F')
		showFps = !showFps;
	else if (beatenGame && key == 'I')
		invert = !invert;
	else if (helpScreen)
	{
		helpScreen = false;
		return;
	}
	if (activeMenu)
		activeMenu->key(key);
	else if (key == GLFW_KEY_ESC || key == 'P')
		pauseMenu.open();
	else if (world.goalreached && key == GLFW_KEY_ENTER)
	{
		if (curlev + 1 < (signed)levels.size())
			startLevel(curlev + 1);
		else
		{
			world.load("title");
			mainMenu.open();
			if (!beatenGame)
				invert = beatenGame = true;
		}
	}
	else if (key == 'C')
		camera = (camera + 1) % 2;
}

int main(int argc, char* argv[])
{
	int wf = GLFW_FULLSCREEN;
	int w = 800, h = 600;
	for (int i = 1; i < argc; i++)
		if (strcmp(argv[i], "-window") == 0)
			wf = GLFW_WINDOW;
		else if (strcmp(argv[i], "-cpufriendly") == 0)
			cpufriendly = true;
		else if (strcmp(argv[i], "-640") == 0)
		{
			w = 640;
			h = 480;
		}
		else if (strcmp(argv[i], "-1024") == 0)
		{
			w = 1024;
			h = 768;
		}

	// load level list
	ifstream f("lvls/levels.txt");
	while (f.good())
	{
		string s;
		getline(f, s);
		if (!s.empty())
		{
			levels.push_back(s);
			getline(f, s);
			levelnames.push_back(s);
		}
	}
	f.close();

	// sound
	if (FSOUND_Init(44100, 32, 0))
		gotSound = true;

	// set up opengl framework
	if (!glfwInit())
	{
		printf("Unable to init GL framework");
		return 1;
	}

	if (!glfwOpenWindow(w, h, 8, 8, 8, 8, 24, 0, wf))
	{
		printf("Unable to find a suitable screen mode");
		return 1;
	}

	glViewport(0, 0, w, h);

	glfwSetKeyCallback(keyHandler);
	glfwEnable(GLFW_KEY_REPEAT);

	glEnable(GL_CULL_FACE);	
	glEnable(GL_TEXTURE_2D);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	font.FaceSize(16);

	world.load("title");

	const double delay = 1.f / 100.f;
	double nextframe = delay;
	double nextlight = delay * 2;
	world.time = 0;
	world.frame = 0;
	int frames = 0;
	double nextfc = glfwGetTime() + 1;
	double time = glfwGetTime();
	double lasttime = time;
	mainMenu.open();
	while (!exitGame)
	{
		time = glfwGetTime();
		double dt = time - lasttime;
		lasttime = time;
		if (dt > .05)
			dt = .05;
		if (activeMenu && activeMenu != &mainMenu || helpScreen)
			dt = 0;

		nextframe -= dt;
		nextlight -= dt;
		while (nextframe <= 0 || nextlight <= 0)
		{
			if (nextframe <= 0)
			{
				nextframe += delay;
				world.time += delay;
				world.frame++;
				world.update();
			}
			if (nextframe <= 0)
				glfwPollEvents();

			if (nextlight <= 0)
			{
				nextlight += delay * 2;
				world.doLights();
			}
		}

		glDepthMask(1);
		glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);

		setupForWorld();
		if (camera == 1 && world.player)
			gluLookAt(world.player->x, 23, world.player->y - 6,
			          world.player->x, 0, world.player->y - 1,
			          0, 1, 0);
		else
			gluLookAt(world.width / 2, 30, world.height / 2 - 18,
			          world.width / 2, 0, world.height / 2 - 3,
			          0, 1, 0);
		if (!helpScreen)
			world.render();

		setupForHUD();
		if (helpScreen)
		{
			static GLuint hlp = 0;
			if (!hlp)
			{
				glGenTextures(1, &hlp);
				glBindTexture(GL_TEXTURE_2D, hlp);
				if (!glfwLoadTexture2D("gfx/help.tga", 0))
					printf("could not load help screen");
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			}
			glBindTexture(GL_TEXTURE_2D, hlp);
			glColor3f(1.f, 1.f, 1.f);
			glDisable(GL_BLEND);
			glBegin(GL_QUADS);
			glTexCoord2i(0, 0);
			glVertex2i(400 - 256, 300 - 256);
			glTexCoord2f(1, 0);
			glVertex2i(400 + 256, 300 - 256);
			glTexCoord2f(1, 1);
			glVertex2i(400 + 256, 300 + 256);
			glTexCoord2f(0, 1);
			glVertex2i(400 - 256, 300 + 256);
			glEnd();
		}
		else
		{
			if (!activeMenu || activeMenu == &pauseMenu)
				drawGameHUD();
			if (activeMenu)
				activeMenu->draw();
		}

		glColor3f(1.f, 1.f, 1.f);

		// fps shit
		frames++;
		static char buffer[64] = "Fps: n/a";
		if (nextfc <= glfwGetTime())
		{
			sprintf(buffer, "Fps: %d", frames);
			frames = 0;
			nextfc = glfwGetTime() + 1;
		}
		if (showFps)
			print(font, 4.f, 600 - 34, buffer);

		//
		if (activeMenu == &mainMenu)
		{
			if (beatenGame)
				print(font, 400.f, 28.f, "Toggle invert mode with I.", true, true);
			print(font, 400.f, 6.f, "Uplighter by Jolle in 2005. Written in 48h for the Ludum Dare 48h game programming conest.", false, true);
		}

		// invert ;)
		if (invert)
		{
			glEnable(GL_BLEND);
			glBlendFunc(GL_ONE_MINUS_DST_COLOR, GL_ZERO);
			glDisable(GL_TEXTURE_2D);
			glColor3f(1.f, 1.f, 1.f);
			glBegin(GL_QUADS);
			glVertex2i(0, 0);
			glVertex2i(800, 0);
			glVertex2i(800, 600);
			glVertex2i(0, 600);
			glEnd();
			glEnable(GL_TEXTURE_2D);
		}

		glfwSwapBuffers();
		if (cpufriendly)
			glfwSleep(.001);
	}

	glfwTerminate();

	if (gotSound)
		FSOUND_Close();

	return 0;
}

void startLevel(int n)
{
	curlev = n;
	world.load(levels[n]);
	activeMenu = 0;
}

void setupForWorld()
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.f, 4.f / 3.f, 1.f, 1000.f);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glEnable(GL_DEPTH_TEST);
}

void setupForHUD()
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, 800, 0, 600, -1, 1);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glDisable(GL_DEPTH_TEST);
}

void drawGameHUD()
{
	if (!world.player)
		return;

	glDisable(GL_BLEND);

	// percent shower
	int x2 = 4 * (int)(200 * world.percent + .5);
	int x3 = 8 * world.targetpercent;
	if (x3 < x2)	
		x3 = x2;
	drawBar(0, x2, x3, 800, 0, 24);
	glColor3f(1.f, 1.f, 1.f);
	print(font, 4.f, 6.f, toStr(int(world.percent * 100 + .25f)) + "% lit, " + toStr(world.targetpercent) + "% needed.", true);

	// "ammo" shower
	const int yplace = 24;
	if (world.player->nrLights)
		drawBar(0, 100 - 4 * (world.player->lightsCountdown / 4), 100, 100, yplace + 24, 24);
	else
		drawBar(0, 0, 100, 100, yplace + 24, 24);
	glColor3f(1.f, 1.f, 1.f);
	print(font, 4.f, yplace + 30, "Lights: " + toStr(world.player->nrLights), true);

	if (world.player->nrBreakers)
		drawBar(0, 100 - 4 * (world.player->breakerCountdown / 4), 100, 100, yplace, 24);
	else
		drawBar(0, 0, 100, 100, yplace, 24);
	glColor3f(1.f, 1.f, 1.f);
	print(font, 4.f, yplace + 6, "Bombs: " + toStr(world.player->nrBreakers), true);

	// ready string?
	if (world.goalreached)
		if (curlev + 1 == (signed)levels.size())
		{
			print(font, 400.f, 600 - 48, "Congratulations!", true, true);
			if (!beatenGame)
			{
				print(font, 400.f, 600 - 64, "You've completed the game!", true, true);
				print(font, 400.f, 600 - 96, "The skip level cheat is F1+F11.", true, true);
				print(font, 400.f, 600 - 114, "Invert mode will be activated :)", true, true);
			}
			else
				print(font, 400.f, 600 - 96, "You've completed the game... again!", true, true);
			print(font, 400.f, 600 - 146, "Press Enter to continue.", true, true);
		}
		else
			print(font, 400.f, 600 - 34, "Goal reached, press Enter to move on.", true, true);
	print(font, 4.f, 600 - 18, "Level " + toStr(curlev + 1) + ": " + levelnames[curlev], true);
}

void drawBar(int x1, int x2, int x3, int x4, int y1, int h)
{
	static GLuint prog = 0;
	if (!prog)
	{
		glGenTextures(1, &prog);
		glBindTexture(GL_TEXTURE_2D, prog);
		glfwLoadTexture2D("gfx/grad.tga", 0);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}

	glBindTexture(GL_TEXTURE_2D, prog);
	glColor3f(0.f, .9f, 0.f);
	glBegin(GL_QUADS);
	int lw = x1;
	int w = x2;
	glTexCoord2f(lw * .25f, 0);
	glVertex2i(lw, y1);
	glTexCoord2f(w * .25f, 0);
	glVertex2i(w, y1);
	glTexCoord2f(w * .25f, 1);
	glVertex2i(w, y1 + h);
	glTexCoord2f(lw * .25f, 1);
	glVertex2i(lw, y1 + h);
	lw = w;
	w = x3;
	glColor3f(.9f, 0.f, 0.f);
	glTexCoord2f(lw * .25f, 0);
	glVertex2i(lw, y1);
	glTexCoord2f(w * .25f, 0);
	glVertex2i(w, y1);
	glTexCoord2f(w * .25f, 1);
	glVertex2i(w, y1 + h);
	glTexCoord2f(lw * .25f, 1);
	glVertex2i(lw, y1 + h);
	lw = w;
	w = x4;
	glColor3f(.9f, .9f, 0.f);
	glTexCoord2f(lw * .25f, 0);
	glVertex2i(lw, y1);
	glTexCoord2f(w * .25f, 0);
	glVertex2i(w, y1);
	glTexCoord2f(w * .25f, 1);
	glVertex2i(w, y1 + h);
	glTexCoord2f(lw * .25f, 1);
	glVertex2i(lw, y1 + h);
	glEnd();
}

void print(FTGLTextureFont& fnt, float x, float y, const string& str, bool shadow, bool center)
{
	if (center)
	{
		float x1, x2, w;
		fnt.BBox(str.c_str(), x1, w, w, x2, w, w);
		w = x2 - x1;
		x -= w * .5;
	}
	glPushMatrix();
	glLoadIdentity();
	if (shadow)
	{
		glTranslatef(x + 1.f, y - 1.f, 0.f);
		glPushAttrib(GL_CURRENT_BIT);
		glColor3f(0.f, 0.f, 0.f);
		fnt.Render(str.c_str());
		glPopAttrib();
		glLoadIdentity();
	}
	glTranslatef(x, y, 0.f);
	fnt.Render(str.c_str());
	glPopMatrix();
}

void playSfx(const string& snd)
{
	if (!gotSound)
		return;

	static map<string, FSOUND_SAMPLE*> samples;

	FSOUND_SAMPLE* smpl = 0;
	map<string, FSOUND_SAMPLE*>::iterator s = samples.find(snd);
	if (s != samples.end())
		smpl = s->second;
	else
	{
		smpl = FSOUND_Sample_Load(FSOUND_FREE, snd.c_str(), FSOUND_NORMAL, 0, 0);
		samples[snd] = smpl;
	}

	if (smpl)
		FSOUND_PlaySound(FSOUND_FREE, smpl);
}
